# -*- coding: utf-8 -*-

from django.core.management.base import BaseCommand, CommandError
from django.db import transaction
from django.db.models import Q
from django.utils.http import parse_http_date_safe

from icalendar import Calendar
import os
import requests
import re
import sys
from portail import settings

from edt.backends.edt_ade import _set_timetable, _open_edt_ade_shelf
from edt.models import EdtAdeGroup
from edt.timetable import TimetableFactory, Category
from portailuser.models import Ecole

# https://urllib3.readthedocs.org/en/latest/security.html#insecurerequestwarning
from requests.packages import urllib3

urllib3.disable_warnings()


AUTHENTICATION = {
    Ecole.n7: settings.EDT_AUTH.get("n7", (None, None)),
}
PROJECT_ID = {
    Ecole.a7: settings.EDT_PROJECT_ID.get("a7", None),
    Ecole.ensat: settings.EDT_PROJECT_ID.get("ensat", None),
    Ecole.n7: settings.EDT_PROJECT_ID.get("n7", None),
}
HTTP_TIMEOUT = 30
NAME_FIXUPS = {
    (500, "Mécaflu groupe1 1A"): "Mécaflu groupe1 2A",
    (501, "Mécaflu groupe2 1A"): "Mécaflu groupe2 2A",
}

GROUPS_LIST_URL = "https://edt.inp-toulouse.fr/direct/gwtdirectplanning/"
GROUP_URL = (
    "https://edt.inp-toulouse.fr/jsp/custom/modules/plannings/"
    "anonymous_cal.jsp?resources={}&projectId=1&calType=ical&nbWeeks=30"
)
GROUPS_LIST_URL = "https://edt.inp-toulouse.fr/direct/gwtdirectplanning/"

VACANCES = {
    "Toussain",
    "Libre (vacances)",
    "Pâques",
    "",
}


def _gen_full_group(ident, profondeur, obj):
    """Génération de la chaine de caractère pour accéder à la ressource"""
    res = '{"%s""true""%i""-1""0""0""0""false"[%i]' % (
        ident,
        min(0, profondeur - 1),
        profondeur,
    )
    res += "{" + obj
    if profondeur > 0:
        res += '1""0"[0][0]'
    return res


class ADEGroupListAccessor(object):
    HEADERS = {
        "Content-Type": "text/x-gwt-rpc; charset=UTF-8",
        "X-GWT-Module-Base": "https://edt.inp-toulouse.fr/direct/gwtdirectplanning/",
    }
    SERVICES = {
        "WebClient": (
            "4119696D5F69F9E2EBD4DE38D3F55A20",
            "com.adesoft.gwt.core.client.rpc.WebClientServiceProxy",
        ),
        "DirectPlanning": (
            "D299C8C3CA21CA5E6AFCED14CFFB2A29",
            "com.adesoft.gwt.directplan.client.rpc.DirectPlanningServiceProxy",
        ),
    }

    TYPE_LOGINREQUEST = "com.adesoft.gwt.core.client.rpc.data.LoginRequest/" "3705388826"
    TYPE_DIRECTLOGINREQUEST = "com.adesoft.gwt.directplan.client.rpc.data." "DirectLoginRequest/635437471"
    TYPE_TREERESCONFIG = "com.adesoft.gwt.directplan.client.ui.tree." "TreeResourceConfig/2234901663"

    BASE_GROUP = '{"-100""true""-1""-1""-1""-1""0""false"[0]"""""0""0"[0][0]'

    class RPCError(Exception):
        pass

    class CommandError(Exception):
        pass

    def __init__(self, username, password):
        self._session = requests.Session()
        self._session.headers.update(self.HEADERS)
        self._ident = "aaaaaab"  # Devrait changer à chaque requête
        self._folder_names = set()

        self.cookie = self._login(username, password)
        self._load_project(self._get_last_project_id())
        self._items = self._get_leaves(self.BASE_GROUP)

    def __iter__(self):
        return iter(self._items)

    def get_folder_names(self):
        return self._folder_names

    def _login(self, login, password):
        # On récupère le header X-GWT-Permutation
        header_text = self._session.get(
            "https://edt.inp-toulouse.fr/direct/gwtdirectplanning/gwtdirectplanning.nocache.js",
            verify=False,
        ).text
        m = re.search(".*,cc='(?P<permutation>[A-F0-9]+)'.*", header_text)
        self._session.headers.update({"X-GWT-Permutation": m.group("permutation")})

        result = self._do_request(
            "WebClient",
            "method1login",
            9,
            [
                "J",
                self.TYPE_LOGINREQUEST,
                self.TYPE_DIRECTLOGINREQUEST,
                login,
                password,
                1,
                2,
                3,
                4,
                2,
                5,
                6,
            ],
            [
                7,
                0,
                0,
                0,
                0,
                0,
                8,
                9,
                -1,
                0,
                0,
            ],
        )

        if "MsgInvalidLogin" in result:
            raise self.CommandError("Invalid login/password")

    def _get_last_project_id(self):
        result = self._do_request("WebClient", "method4getProjectList", 5, ["J", 1, 2, 3, 4, 1, 5], [])

        m = re.search('\\"(?P<id>[0-9]+)', result)
        return m.group("id")

    def _load_project(self, project_id):
        result = self._do_request(
            "DirectPlanning",
            "method13loadProject",
            7,
            [
                "J",
                "I",
                "Z",
                1,
                2,
                3,
                4,
                3,
                5,
                6,
                7,
            ],
            [
                project_id,
                1,
            ],
        )

        if result != "[[],0,7]":
            raise self.CommandError("Loading project failed?")

    INFO_RE = re.compile(r'"StringField""NAME""LabelName""[^"]*""(true|false)""false""""([^"]*?)""')

    def _get_leaves(self, group_obj, depth=0):
        leaves = []

        subgroups = self._get_subgroups(group_obj)
        for ident, is_folder, name, sub_obj in subgroups[depth:]:
            if is_folder and "DETTES" in name:
                continue

            if is_folder:
                self._folder_names.add(name)
                group_leaves = self._get_leaves(_gen_full_group(ident, depth + 1, sub_obj), depth + 1)
                leaves += group_leaves
            else:
                # Fix the name up if necessary
                name = NAME_FIXUPS.get((ident, name), name)
                leaves.append((ident, name))

        print(leaves)
        return leaves

    def _get_subgroups(self, group_obj):
        assert group_obj
        result = self._do_request(
            "DirectPlanning",
            "method4getChildren",
            20,
            [
                "J",
                "java.lang.String/2004016611",
                self.TYPE_TREERESCONFIG,
                group_obj,
                "[I/2970817851",
                "java.util.LinkedHashMap/3008245022",
                "COLOR",
                "com.adesoft.gwt.core.client.rpc.config.OutputField/870745015",
                "LabelColor",
                "",
                "com.adesoft.gwt.core.client.rpc.config.FieldType/1797283245",
                "NAME",
                "LabelName",
                "java.util.ArrayList/4159755760",
                "com.extjs.gxt.ui.client.data.SortInfo/1143517771",
                "com.extjs.gxt.ui.client.Style$SortDir/3873584144",
                1,
                2,
                3,
                4,
                3,
                5,
                6,
                7,
            ],
            [
                8,
                7,
                0,
                9,
                2,
                -1,
                -1,
                10,
                0,
                2,
                6,
                11,
                12,
                0,
                13,
                11,
                14,
                15,
                11,
                0,
                0,
                6,
                16,
                12,
                0,
                17,
                16,
                14,
                15,
                4,
                0,
                0,
                18,
                0,
                18,
                0,
                19,
                20,
                1,
                16,
                18,
                0,
            ],
        )

        data = result[4:-6]
        data = data.replace("\\", "").split("{")
        print(data)

        results = []
        i = 1
        while i < len(data):
            match = self.INFO_RE.search(data[i])
            if match:
                previous_line = data[i - 1].split('"')
                ident = previous_line[1]
                is_folder = previous_line[3] == "true"
                name = match.group(2)
                print(name)
                group_obj = match.group(0)
                results.append((ident, is_folder, name, group_obj))
            i += 1

        return results

    def _do_request(self, service, name, count, args_before, args_after):
        srv_ident, srv_class = self.SERVICES[service]

        items = [7, 0, count, GROUPS_LIST_URL, srv_ident, srv_class, name]
        items += args_before
        items.append(self._ident)
        items += args_after

        data = "|".join(str(i) for i in items) + "|"

        response = self._session.post(
            GROUPS_LIST_URL + service + "ServiceProxy",
            data=data,
            timeout=HTTP_TIMEOUT,
            verify=False,
        )

        response.raise_for_status()

        if response.text.startswith("//OK"):
            return response.text[4:]

        raise self.RPCError(response.text)


class ADEEventExtractor(object):
    """
    A generator of properly formatted events from an ADE icalendar file data.
    """

    _match_auto = None
    _match_manual = None

    class ADEEvent(object):
        __slots__ = (
            "uid",
            "dtstart",
            "dtend",
            "topic",
            "type",
            "category",
            "locations",
            "people",
        )

        def __init__(self):
            for item in self.__slots__:
                setattr(self, item, None)

        def __repr__(self):
            res = ["<ADEEvent: "]
            for i, item in enumerate(self.__slots__):
                if i > 0:
                    res.append(", ")
                res.append(f"{item}={repr(getattr(self, item))}")
            res.append(">")
            return "".join(res)

    class BadEventError(Exception):
        pass

    def __init__(self, data, grouplist, remaining_errors):
        self._calendar = Calendar.from_ical(data)
        self._iter = iter(self._calendar.walk())
        next(self._iter)
        self._grouplist = grouplist
        self.remaining_errors = remaining_errors

        if self.__class__._match_auto is None:
            self.__class__._match_auto = match_auto = []

            with open(os.path.join(os.path.dirname(__file__), "data", "match_auto.txt")) as f:
                for line in f:
                    line = line.strip()
                    if line:
                        match_auto.append(re.compile(line))

            self.__class__._match_manual = match_manual = []
            with open(os.path.join(os.path.dirname(__file__), "data", "match_manual.txt")) as f:
                for line in f:
                    line = line.rstrip()
                    if not line:
                        continue

                    regex, type_text, topic, cat_text = re.split(r"\t+", line)
                    regex = re.compile(regex)
                    if type_text == ".":
                        type_text = None
                    else:
                        type_text = type_text
                    topic = topic
                    cat = getattr(Category, cat_text)

                    match_manual.append((regex, type_text, topic, cat))

    def __iter__(self):
        return self

    def __next__(self):
        """
        Get the next formatted event object from the iterator
        """

        event = self.ADEEvent()

        while True:
            cur = next(self._iter)

            try:
                ok = self._fill_event(event, cur)
            except self.BadEventError as err:
                print(f"Unable to parse the following event ({err}) :")
                print(f" UID:         {cur['UID']}")
                print(f" Start date:  {cur['DTSTART'].dt}")
                print(f" End date:    {cur['DTEND'].dt}")
                print(f" Location:    {cur['LOCATION']}")
                print(f" Summary:     '{str(cur['SUMMARY'])}'")
                print(f" Description: '{str(cur['DESCRIPTION'])}'")
                print(f" Parsed now:  {repr(event)}")

                self.remaining_errors -= 1

                if self.remaining_errors < 1:
                    sys.exit(1)

                ok = False

            if ok:
                return event

    def _fill_event(self, event, cur):
        """
        Fills the event with the icalendar data in cur.
        Returns True if the event is to be stored
        """

        event.uid = str(cur["UID"])
        event.dtstart = cur["DTSTART"].dt
        event.dtend = cur["DTEND"].dt
        event.locations = [item for item in cur["LOCATION"].split(",") if item]

        desc = cur["DESCRIPTION"].split("\n")
        # Fake descriptions start with an empty line and have the list of
        # groups followed by the list of people followed by an add date.
        # Real description do not have the empty line, apparently
        if desc[0]:
            event.people = []
        else:
            event.people = [
                item for item in desc if item and item not in self._grouplist and not item.startswith("(Exporté le:")
            ]

        summary = str(cur["SUMMARY"])

        if summary in VACANCES:
            return

        for regex in self._match_auto:
            res = regex.match(summary)
            if res:
                type_text = res.group("cat")
                topic = res.group("topic")
                if topic.startswith(type_text + " "):
                    topic = topic[len(type_text) + 1 :]
                return self._fill_event_cat(event, type_text, topic)

        for regex, type_text, topic, category in self._match_manual:
            res = regex.match(summary)
            if res:
                event.type = type_text
                event.topic = topic
                event.category = category
                return True

        event.type = "Mal formaté sur ADE"
        event.topic = summary
        event.category = "AUTRE"

        return True
        # raise self.BadEventError(f"Unknown summary {[repr(summary)}")

    def _fill_event_cat(self, event, type_text, topic):
        """
        Extract info from type text and topic
        """

        type_text = type_text.strip()
        topic = topic.strip()

        event.topic = topic

        if type_text in {"C", "CM", "CMM"}:
            event.type = "Cours"
            event.category = Category.COURS

        elif type_text in {"TP", "TPE"}:
            event.type = type_text
            event.category = Category.TP

        elif type_text in {"TD", "CTD", "TDM", "CM TD"}:
            event.type = type_text
            event.category = Category.TD

        elif type_text in {"EXAM", "Examen", "EXAMEN", "TEST"}:
            event.type = "Examen"
            event.category = Category.EXAMEN

        elif type_text in {"ORAL", "ORA", "Oral"}:
            event.type = "Oral"
            event.category = Category.EXAMEN

        elif type_text in {"Projet", "BE", "TPBE", "TPPR"}:
            event.type = type_text
            event.category = Category.PROJET

        elif type_text == "VISITE":
            event.type = "Visite"
            event.category = Category.AUTRE

        elif type_text in {"Conf", "Conférences"}:
            event.type = "Conférence"
            event.category = Category.COURS

        elif type_text == "TUTORAT":
            event.type = "Tutorat"
            event.category = Category.AUTRE

        elif type_text == "EVALPROJ":
            event.type = "Éval. projet"
            event.category = Category.EXAMEN

        elif type_text == "Présentation":
            event.type = type_text
            event.category = Category.AUTRE

        elif type_text == "rattrapage":
            event.type = "Rattrapages"
            event.category = Category.EXAMEN

        else:
            raise self.BadEventError(f"Unknown type text {repr(type_text)} (topic {repr(topic)})")

        return True


class Command(BaseCommand):
    TOPIC_RE_1 = re.compile(r"^(.*) \(\1\)$")
    TOPIC_RE_2 = re.compile(r"^(.*) \(.*?\)$")

    def __init__(self):
        super(BaseCommand, self).__init__()
        self._verbosity = 0
        self._shelf = None
        self._remaining_errors = 1

    def add_arguments(self, parser):
        parser.add_argument(
            "--no-update-group-list",
            action="store_false",
            dest="update_list",
            default=True,
            help="Prevent updating the group list",
        )
        parser.add_argument(
            "--start-at",
            action="store",
            type=int,
            dest="start_at",
            default=1,
            help="Update starting at the specified sequential group number",
        )
        parser.add_argument(
            "--max-errors",
            action="store",
            type=int,
            dest="max_errors",
            default=1,
            help="Max number of errors before aborting",
        )

    def handle(self, *args, **options):
        self._verbosity = int(options.pop("verbosity", "0"))
        self._remaining_errors = options["max_errors"]
        update_list = options.pop("update_list", True)
        start_at = options["start_at"]

        with _open_edt_ade_shelf() as shelf:
            self._shelf = shelf

            for ecole, (username, password) in AUTHENTICATION.items():
                if update_list:
                    groups, names = self._update_group_list(ecole, username, password)
                else:
                    groups = EdtAdeGroup.objects.filter(ecole=ecole)
                    names = set(groups.values_list("name", flat=True))
                    groups = groups.exclude(gid=None)

                count = len(groups)
                for i, group in enumerate(groups):
                    if i + 1 < start_at:
                        continue

                    self._debug(f"[{i+1:{len(str(count))}d}/{i+1}] Updating group id {count} ({group.gid})")
                    self._update_group(group, PROJECT_ID[ecole], names)

    @transaction.atomic
    def _update_group_list(self, ecole, username, password):
        self._debug("Updating group list for {} (will take some time...)", ecole)

        cur_groups = list(EdtAdeGroup.objects.filter(ecole=ecole).exclude(gid=None))
        cur_group_ids = set(group.gid for group in cur_groups)
        new_groups = []
        new_group_ids = set()
        group_names = set()

        accessor = ADEGroupListAccessor(username, password)
        for gid, name in accessor:
            if isinstance(gid, str):
                gid = int(gid)
            group_names.add(name)
            if gid in new_group_ids:  # Already seen (?)
                continue

            new_group_ids.add(gid)

            if gid in cur_group_ids:
                cur_group_ids.remove(gid)
            else:
                group = EdtAdeGroup(ecole=ecole, gid=gid, name=name)
                new_groups.append(group)
                cur_groups.append(group)

        added_count = len(new_groups)
        removed_count = len(cur_group_ids)

        for folder_name in accessor.get_folder_names():
            if folder_name not in group_names:
                new_groups.append(EdtAdeGroup(ecole=ecole, gid=None, name=folder_name))
                group_names.add(folder_name)

        EdtAdeGroup.objects.filter(Q(ecole=ecole) & (Q(gid__in=cur_group_ids) | Q(gid=None))).delete()
        EdtAdeGroup.objects.bulk_create(new_groups)

        if added_count or removed_count:
            # Always display this so we know if the group list accessor
            # messed up
            print(f"Group list updated, {added_count} added, {removed_count} removed.")
        else:
            self._debug("Group list updated, no changes.")

        return sorted(cur_groups, key=lambda item: item.gid), group_names

    def _update_group(self, group, school_id, names):
        gid = group.gid

        response = requests.get(GROUP_URL.format(gid), timeout=HTTP_TIMEOUT, verify=False)
        response.raise_for_status()

        content_type = response.headers["content-type"].replace(" ", "")

        factory = TimetableFactory(group.name)
        data = response.content
        if content_type == "text/calendar;charset=UTF-8":
            extractor = ADEEventExtractor(data, names, self._remaining_errors)
            for event in extractor:
                factory.add_event(
                    event.dtstart,
                    event.dtend,
                    event.topic,
                    event.type,
                    event.category,
                    event.locations,
                    event.people,
                    event.uid,
                )
            self._remaining_errors = extractor.remaining_errors
        elif content_type != "text/html;charset=ISO-8859-1":
            raise CommandError(f"GID {gid} has bad content type {content_type}")

        timetable = factory.table
        timetable.timestamp = parse_http_date_safe(response.headers.get("Last-Modified"))
        _set_timetable(self._shelf, school_id, gid, factory.table)

    def _debug(self, msg, *args):
        if self._verbosity < 1:
            return

        print(msg.format(args) + "\n")
